/* Copyright (c) 2022 Skyward Experimental Rocketry
 * Author: Federico Mandelli
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

#pragma once

#include <ActiveObject.h>
#include <utils/Debug.h>
#include <utils/collections/IRQCircularBuffer.h>

#include <thread>

#include "Canbus.h"

#define N_BOARDS 3  ///< Number of boards on the bus.

namespace Boardcore
{

namespace Canbus
{

/**
 * The id of a can packet is composed of 29 bits and will be divided such as:
 * - Priority           4 bit - priority    \
 * - Type               6 bit - type        |
 * - Source             4 bit - source      | 22 bits
 * - Destination        4 bit - destination |
 * - Type id            4 bit - idType      /
 * - First packet flag  1 bit - firstPacket \ 7 bits
 * - Remaining packets  6 bit - leftToSend  /
 * shiftNameOfField the number of shift needed to reach that field
 */

/**
 * @brief The mask of the ID without the sequential information.
 *
 * CompleteID = (IDMask << shiftSequentialInfo) || SequentialInformation
 */

/**
 * @brief Enumeration that contains masks of the elements composing the can
 * packet id without sequential information.
 */
enum IDMask : uint32_t
{
    priority = 0x3C0000,

    type        = 0x03F000,
    source      = 0x000F00,
    destination = 0x0000F0,
    idType      = 0x00000F
};

/**
 * @brief Enumeration that contains masks of the elements composing the
 * sequential information.
 */
enum SequentialInformation : uint8_t
{
    firstPacket = 0x40,
    leftToSend  = 0x3F
};

enum ShiftInformation : uint8_t
{
    // Shift info of IDMask
    shiftPriority    = 18,
    shiftType        = 12,
    shiftSource      = 8,
    shiftDestination = 4,
    shiftIdType      = 0,
    // Shift info of SequentialInformation
    shiftFirstPacket    = 6,
    shiftLeftToSend     = 0,
    shiftSequentialInfo = 7
};

/**
 * @brief Generic struct that contains a logical can packet.
 *
 * i.e. 1 accelerometer packet 3 * 4byte (acc: x,y,z)  +timestamp, will be 4
 * canPacket but a single canData.
 */
struct CanData
{
    int32_t canId = -1;  ///< Id of the packet without the sequential info.
    uint8_t length;
    uint8_t nRec = 0;
    uint64_t payload[32];
};

/**
 * @brief Canbus protocol implementation.
 *
 * Given a can interface this class takes care of sending CanData packets
 * segmented into multiple CanPackets and receiving single CanPackets that are
 * then reframed as CanData.
 */
class CanProtocol : public ActiveObject
{
    using MsgHandler = std::function<void(CanData data)>;

public:
    /**
     * @brief Construct a new CanProtocol object.
     *
     * @param can Pointer to a CanbusDriver object.
     */
    explicit CanProtocol(CanbusDriver* can, MsgHandler callback);

    ~CanProtocol();

    /**
     * @brief Sends a CanData object on the bus.
     *
     * Takes a CanData object, splits it into multiple CanPackets with the
     * correct sequential id.
     * @warning requires @param data to be not empty.
     *
     * @param data Contains the id and the data of the packet to send.
     */
    void send(CanData toSend);

private:
    /**
     * @brief Count the number of bytes needed to encode a uint64_t number.
     */
    uint8_t byteForInt(uint64_t number);

    void sendData();

    /**
     * @brief Keeps listening on the canbus for packets.
     *
     * Once a packet is received, it checks if it is expected (that id is
     * already present in data), if it is the case, it's added to the list.
     * Once we receive the correct amount of packet we offer it in buffer.
     *
     * For now if a packet is missed/received in the wrong order the whole
     * packet will be lost once we receive a new first packet without warning
     * CanHandler.
     */
    void run() override;

    // The physical implementation of the CanBus
    CanbusDriver* can;

    // The buffer used to store the CanData to send
    IRQCircularBuffer<CanData, N_BOARDS> TXbuffer;

    miosix::FastMutex mutex;

    MsgHandler callback;
};

}  // namespace Canbus

}  // namespace Boardcore
